import {
  AggregateID,
  ConflictException,
  MessagePublisherPort,
} from '@mobicoop/ddd-library';
import {
  AD_MESSAGE_PUBLISHER,
  AD_REPOSITORY,
  AD_ROUTE_PROVIDER,
} from '@modules/ad/ad.di-tokens';
import { AdAlreadyExistsException } from '@modules/ad/core/domain/ad.errors';
import { AdFactory } from '@modules/ad/core/domain/ad.factory';
import { Inject } from '@nestjs/common';
import { CommandHandler, ICommandHandler } from '@nestjs/cqrs';
import { MATCHER_AD_CREATION_FAILED_ROUTING_KEY } from '@src/app.constants';
import { GeorouterService } from '../../../domain/georouter.service';
import { MatcherAdCreationFailedIntegrationEvent } from '../../events/matcher-ad-failure.integration-event';
import { AdRepositoryPort } from '../../ports/ad.repository.port';
import { CreateAdCommand } from './create-ad.command';

@CommandHandler(CreateAdCommand)
export class CreateAdService implements ICommandHandler {
  constructor(
    @Inject(AD_MESSAGE_PUBLISHER)
    private readonly messagePublisher: MessagePublisherPort,
    @Inject(AD_REPOSITORY)
    private readonly repository: AdRepositoryPort,
    @Inject(AD_ROUTE_PROVIDER)
    private readonly routeProvider: GeorouterService,
  ) {}

  async execute(command: CreateAdCommand): Promise<AggregateID> {
    try {
      const adFactory = new AdFactory(this.routeProvider);
      const ad = await adFactory.create(command);

      try {
        //TODO it should not be this service's concern that Prisma does not support postgis types
        await this.repository.insertExtra(ad, 'ad');
        return ad.id;
      } catch (error: any) {
        if (error instanceof ConflictException) {
          throw new AdAlreadyExistsException(error);
        }
        throw error;
      }
    } catch (error: any) {
      const matcherAdCreationFailedIntegrationEvent =
        new MatcherAdCreationFailedIntegrationEvent({
          id: command.id,
          metadata: {
            correlationId: command.id,
            timestamp: Date.now(),
          },
          cause: error.message,
        });
      this.messagePublisher.publish(
        MATCHER_AD_CREATION_FAILED_ROUTING_KEY,
        JSON.stringify(matcherAdCreationFailedIntegrationEvent),
      );
      throw error;
    }
  }
}
